<?php
/**
 * @package     RO CSVI
 * @subpackage  RO CSVI Breakdesigns Custom fields
 *
 * @author      RolandD Cyber Produksi <contact@rolandd.com>
 * @copyright   Copyright (C) 2006 - 2015 RolandD Cyber Produksi. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 * @link        https://rolandd.com
 */

defined('_JEXEC') or die;

use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Filesystem\File;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\PluginHelper;

/**
 * Process Breakdesigns custom fields.
 *
 * @package     CSVI
 * @subpackage  Plugin.Customfieldsforall
 * @since       6.0
 */
class PlgCsviextCustomfieldsforall extends RantaiPluginDispatcher
{
	/**
	 * JDatabase handler
	 *
	 * @var    JDatabase
	 * @since  6.0
	 */
	protected $db;

	/**
	 * List of parent IDs which are used as derived product
	 *
	 * @var    array
	 * @since  6.5.0
	 */
	private $derivedProducts = [];

	/**
	 * List of product IDs already updated
	 *
	 * @var    array
	 * @since  6.5.0
	 */
	private $productIds = [];

	/**
	 * The VirtueMart configuration
	 *
	 * @var    Com_VirtuemartHelperCom_Virtuemart_Config
	 * @since  7.21.0
	 */
	private $virtuemartConfig;

	/**
	 * Set if this plugin can be processed
	 *
	 * @var    boolean
	 * @since  8.0.0
	 */
	private $canRun;

	/**
	 * Construct the class.
	 *
	 * @since   6.0
	 */
	public function __construct()
	{
		$this->canRun = PluginHelper::isEnabled('csviaddon', 'virtuemart') && ComponentHelper::isEnabled('com_virtuemart');

		if (!$this->canRun)
		{
			return;
		}

		$this->db = Factory::getDbo();

		$jlang = Factory::getLanguage();
		$jlang->load('plg_csviext_customfieldsforall', __DIR__, 'en-GB', true);
		$jlang->load('plg_csviext_customfieldsforall', __DIR__, $jlang->getDefault(), true);
		$jlang->load('plg_csviext_customfieldsforall', __DIR__, null, true);

		if (File::exists(JPATH_PLUGINS . '/csviaddon/virtuemart/com_virtuemart/helper/com_virtuemart_config.php'))
		{
			require_once JPATH_PLUGINS . '/csviaddon/virtuemart/com_virtuemart/helper/com_virtuemart_config.php';
			$this->virtuemartConfig = new Com_VirtuemartHelperCom_Virtuemart_Config;
		}
	}

	/**
	 * Get the main custom Id for stockablecustomfields plugin.
	 *
	 * @param   int  $product_parent_id  The ID of the product parent to check.
	 *
	 * @return  int  Virtuemart custom id.
	 *
	 * @since   6.5.0
	 */
	private function getMainCustomId($product_parent_id)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('c.virtuemart_custom_id'))
			->from($this->db->quoteName('#__virtuemart_customs', 'c'))
			->leftJoin(
				$this->db->quoteName('#__virtuemart_product_customfields', 'pc')
				. ' ON ' . $this->db->quoteName('pc.virtuemart_custom_id') . ' = ' . $this->db->quoteName(
					'c.virtuemart_custom_id'
				)
			)
			->where($this->db->quoteName('c.custom_element') . ' = ' . $this->db->quote('stockablecustomfields'))
			->where($this->db->quoteName('pc.virtuemart_product_id') . ' = ' . (int) $product_parent_id);
		$this->db->setQuery($query);

		return $this->db->loadResult();
	}

	/**
	 * Get the custom fields added for stockablecustomfields plugin.
	 *
	 * @param   int   $mainCustomId  The main ID of the stockablecustomfield.
	 * @param   bool  $onlyQuery     If only query result need to be returned.
	 *
	 * @return  array  Virtuemart custom id array.
	 *
	 * @since   6.5.0
	 */
	private function getCustomFieldIds($mainCustomId, $onlyQuery)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('custom_params'))
			->from($this->db->quoteName('#__virtuemart_customs'))
			->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $mainCustomId);
		$this->db->setQuery($query);
		$result = $this->db->loadResult();

		if ($onlyQuery)
		{
			return $result;
		}

		$cparams = explode('|', $result);

		$customValue = [];

		foreach ($cparams as $cparam)
		{
			if (strpos($cparam, '='))
			{
				// Get the params
				[$name, $value] = explode('=', $cparam);

				if ($name === 'custom_id')
				{
					$customValue = json_decode($value);
				}
			}
		}

		return $customValue;
	}

	/**
	 * Method to get the field options.
	 *
	 * @param   string         $plugin                 The ID of the plugin.
	 * @param   array          $params                 An array of settings.
	 * @param   int            $virtuemart_product_id  The product ID.
	 * @param   int            $virtuemart_custom_id   The custom ID.
	 * @param   CsviHelperLog  $log                    The CSVI logger.
	 *
	 * @return  mixed  String if plugin needs to handle request | False if plugin does not handle request.
	 *
	 * @since   6.0
	 */
	public function getCustomParam($plugin, $params, $virtuemart_product_id, $virtuemart_custom_id, $log)
	{
		if (strpos($plugin, 'customfieldsforall') !== false)
		{
			return '';
		}

		if ($plugin === 'stockablecustomfields')
		{
			return 'parentOrderable=0|custom_id=""|outofstockcombinations="enabled"|child_product_id="0"|';
		}

		return false;
	}

	/**
	 * Method to get the field options.
	 *
	 * @param   string         $plugin                     The ID of the plugin.
	 * @param   string         $params                     An array of settings.
	 * @param   int            $virtuemart_product_id      The product ID.
	 * @param   int            $virtuemart_custom_id       The custom ID.
	 * @param   int            $virtuemart_customfield_id  The custom field ID.
	 * @param   CsviHelperLog  $log                        The CSVI logger.
	 * @param   int            $product_parent_id          The product parent ID.
	 * @param   int            $productAsDerived           The product has to be derived product.
	 * @param   array          $multiLangValues            Multi language value array
	 * @param   string         $parentCustomValue          Parent custom value for dependent custom fields
	 *
	 * @return   mixed|void  Void if plugin needs to handle request | False if plugin does not handle request.
	 *
	 * @since   6.0
	 *
	 * @throws  Exception
	 * @throws  RuntimeException
	 */
	public function onAfterStoreCustomfield(
		string $plugin,
		string $params,
		int $virtuemart_product_id,
		int $virtuemart_custom_id,
		int $virtuemart_customfield_id,
		CsviHelperLog $log,
		int $product_parent_id,
		int $productAsDerived,
		array $multiLangValues = [],
		string $parentCustomValue = ''
	) {
		if (!$this->canRun)
		{
			return;
		}

		if ($product_parent_id > 0 && strpos($plugin, 'customfieldsforall') === false)
		{
			$plugin = $this->usesStockablecustomfields($product_parent_id);
		}

		if ($plugin === 'stockablecustomfields')
		{
			$this->processStockableCustomFields(
				$virtuemart_product_id,
				$virtuemart_custom_id,
				$virtuemart_customfield_id,
				$log,
				$product_parent_id,
				$productAsDerived
			);
		}

		if (strpos($plugin, 'customfieldsforall') !== false)
		{
			$this->processCustomFieldsForAll(
				$params,
				$virtuemart_product_id,
				$virtuemart_custom_id,
				$virtuemart_customfield_id,
				$log,
				$multiLangValues,
				$parentCustomValue
			);
		}

		return true;
	}

	/**
	 * Check if a parent products uses the stockablecustomfields plugin.
	 *
	 * @param   int  $product_parent_id  The ID of the product parent to check.
	 *
	 * @return  string  The name of the used plugin.
	 *
	 * @since   6.5.0
	 */
	private function usesStockablecustomfields($product_parent_id)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('c.custom_value'))
			->from($this->db->quoteName('#__virtuemart_customs', 'c'))
			->leftJoin(
				$this->db->quoteName('#__virtuemart_product_customfields', 'pc')
				. ' ON ' . $this->db->quoteName('pc.virtuemart_custom_id') . ' = ' . $this->db->quoteName(
					'c.virtuemart_custom_id'
				)
			)
			->where($this->db->quoteName('c.field_type') . ' = ' . $this->db->quote('E'))
			->where($this->db->quoteName('c.custom_value') . ' = ' . $this->db->quote('stockablecustomfields'))
			->where($this->db->quoteName('pc.virtuemart_product_id') . ' = ' . (int) $product_parent_id);
		$this->db->setQuery($query);

		return $this->db->loadResult();
	}

	/**
	 * Method to process stockable custom fields.
	 *
	 * @param   int            $virtuemart_product_id      The product ID.
	 * @param   int            $virtuemart_custom_id       The custom ID.
	 * @param   int            $virtuemart_customfield_id  The custom field ID.
	 * @param   CsviHelperLog  $log                        The CSVI logger.
	 * @param   int            $product_parent_id          The product parent ID.
	 * @param   int            $productAsDerived           The product has to be derived product.
	 *
	 * @return   mixed  Void if plugin needs to handle request | False if plugin does not handle request.
	 *
	 * @since   7.13.0
	 *
	 * @throws  Exception
	 * @throws  RuntimeException
	 */
	private function processStockableCustomFields(
		$virtuemart_product_id,
		$virtuemart_custom_id,
		$virtuemart_customfield_id,
		$log,
		$product_parent_id,
		$productAsDerived
	) {
		// Check if the parent has been set as a derived product
		if ($productAsDerived && !in_array($virtuemart_product_id, $this->derivedProducts))
		{
			$mainCustomId      = $this->getMainCustomId($virtuemart_product_id);
			$parentParams      = $this->getCustomFieldIds($mainCustomId, true);
			$parentParamsArray = explode('|', $parentParams);
			$parentOrderValue  = 0;

			foreach ($parentParamsArray as $pparam)
			{
				if (strpos($pparam, '='))
				{
					// Get the params
					list($pname, $pvalue) = explode('=', $pparam);

					if ($pname == 'parentOrderable')
					{
						$parentOrderValue = json_decode($pvalue);
					}
				}
			}

			if ($parentOrderValue)
			{
				// Remove one entry so we have space for inserting new params
				$query = $this->db->getQuery(true)
					->delete($this->db->quoteName('#__virtuemart_product_customfields'))
					->where(
						$this->db->quoteName('customfield_params') . ' LIKE ' . $this->db->quote(
							'%child_product_id="0"%'
						)
					);

				$this->db->setQuery($query, 0, 1)->execute();
				$log->add('One entry removed so to add custom params for current product/parent as derived');

				$this->derivedProducts[] = $virtuemart_product_id;

				$pluginName           = $this->usesStockablecustomfields($virtuemart_product_id);
				$derivedProductParams = 'custom_id=""|child_product_id="' . $virtuemart_product_id . '"|';
				$columns              = [
					'virtuemart_custom_id',
					'customfield_value',
					'virtuemart_product_id',
					'customfield_params',
				];
				$values               = [
					$this->db->quote($mainCustomId),
					$this->db->quote($pluginName),
					$this->db->quote($virtuemart_product_id),
					$this->db->quote($derivedProductParams),
				];
				$query->clear()
					->insert($this->db->quoteName('#__virtuemart_product_customfields'))
					->columns($this->db->quoteName($columns))
					->values(implode(',', $values));

				$this->db->setQuery($query)->execute();
				$log->add('Added current product/parent as derived');
			}
			else
			{
				$log->add('Parent orderable is not set to Yes in custom fields');
			}
		}

		// Check if there are any customfield values and get its ID
		$custom_fields_all_id = $this->getCf4allCustomId($virtuemart_customfield_id);

		// Process is different if its CF4ALL values, so do accordingly
		if ((int) $custom_fields_all_id === (int) $virtuemart_custom_id)
		{
			$fieldValueName  = $this->db->quoteName('customfield_value');
			$fieldParamsName = $this->db->quoteName('customfield_params');

			$query = $this->db->getQuery(true)
				->select('IF(LENGTH(' . $fieldParamsName . ') > 0,' . $fieldParamsName . ', ' . $fieldValueName . ')')
				->from($this->db->quoteName('#__virtuemart_product_customfields'))
				->where($this->db->quoteName('virtuemart_customfield_id') . ' = ' . (int) $virtuemart_customfield_id);
			$this->db->setQuery($query);
			$custom_value = $this->db->loadResult();
			$log->add('Select the custom value from the product custom fields');

			// Call the CF4ALL method to do the rest
			$this->processCustomFieldsForAll(
				$custom_value,
				$virtuemart_product_id,
				$virtuemart_custom_id,
				$virtuemart_customfield_id,
				$log,
				$multiLangValues = [],
				$parentCustomValue = ''
			);
		}

		// Remove unwanted child entries before inserting
		if ($product_parent_id == 0)
		{
			$query = $this->db->getQuery(true)
				->delete($this->db->quoteName('#__virtuemart_product_customfields'))
				->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $virtuemart_custom_id)
				->where($this->db->quoteName('virtuemart_product_id') . ' !=  0')
				->where($this->db->quoteName('customfield_params') . ' = ' . $this->db->quote(''));
			$this->db->setQuery($query)->execute();
			$log->add('Deleted all unwanted entries');
		}

		// Update the parent with the child ID
		if ($product_parent_id > 0 && !in_array($virtuemart_product_id, $this->productIds))
		{
			// Get the main custom field id
			$mainCustomId = $this->getMainCustomId($product_parent_id);
			$params       = $this->getCustomFieldIds($mainCustomId, true);
			$cparams      = explode('|', $params);

			// Clean up the parameter settings
			$newCustomParam = [];

			foreach ($cparams as $cparam)
			{
				if (strpos($cparam, '='))
				{
					[$name, $value] = explode('=', $cparam);

					switch ($name)
					{
						case 'child_product_id':
							$value = '"' . $virtuemart_product_id . '"';
							break;
						case 'custom_id':
							$value = '""';
							break;
					}

					$newCustomParam[] = $name . '=' . $value;
				}
			}

			// Add an empty entry
			$newCustomParam[] = '';

			$fparams = implode("|", $newCustomParam);

			$query = $this->db->getQuery(true)
				->select($this->db->quoteName('virtuemart_customfield_id'))
				->from($this->db->quoteName('#__virtuemart_product_customfields'))
				->where($this->db->quoteName('customfield_params') . ' LIKE ' . $this->db->quote($fparams));
			$this->db->setQuery($query);
			$parent_customfield_id = $this->db->loadResult();

			// Update the parent params and delete the unwanted entries
			if ($parent_customfield_id)
			{
				$query = $this->db->getQuery(true)
					->delete($this->db->quoteName('#__virtuemart_product_customfields'))
					->where(
						$this->db->quoteName('customfield_params') . ' LIKE ' . $this->db->quote(
							'%child_product_id="0"%'
						)
					);
				$this->db->setQuery($query)->execute();
			}
			else
			{
				// Find an empty slot
				$query = $this->db->getQuery(true)
					->select($this->db->quoteName('virtuemart_customfield_id'))
					->from($this->db->quoteName('#__virtuemart_product_customfields'))
					->where(
						$this->db->quoteName('customfield_params') . ' LIKE ' . $this->db->quote(
							'%child_product_id="0"%'
						)
					)
					->where($this->db->quoteName('virtuemart_product_id') . ' = ' . (int) $product_parent_id)
					->order($this->db->quoteName('ordering'));
				$this->db->setQuery($query);

				$parent_customfield_id = $this->db->loadResult();
			}

			if ($parent_customfield_id)
			{
				$query->clear()
					->update($this->db->quoteName('#__virtuemart_product_customfields'))
					->set($this->db->quoteName('customfield_params') . ' = ' . $this->db->quote($fparams))
					->where($this->db->quoteName('virtuemart_customfield_id') . ' = ' . (int) $parent_customfield_id);
				$this->db->setQuery($query)->execute();
				$this->productIds[] = $virtuemart_product_id;
				$log->add('Inserted main custom id for stockable custom fields plugin');

				// Check if there are any rows for disabler or override field to update
				$query->clear()
					->select($this->db->quoteName('virtuemart_customfield_id'))
					->from($this->db->quoteName('#__virtuemart_product_customfields'))
					->where(
						$this->db->quoteName('virtuemart_product_id') . ' = ' . $this->db->quote($product_parent_id)
					)
					->where($this->db->quoteName('customfield_params') . ' != ' . $this->db->quote(''));
				$this->db->setQuery($query);
				$resultParent = $this->db->loadRowList();
				$totalParent  = count($resultParent);

				$query->clear()
					->select(
						$this->db->quoteName(
							[
								'virtuemart_customfield_id',
								'disabler',
								'override',
							]
						)
					)
					->from($this->db->quoteName('#__virtuemart_product_customfields'))
					->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $mainCustomId)
					->where($this->db->quoteName('virtuemart_product_id') . ' = 0')
					->where(
						'(' . $this->db->quoteName('customfield_params') . ' IS NULL OR ' .
						$this->db->quoteName('customfield_params') . ' = ' . $this->db->quote('')
						. ')'
					)
					->where(
						'(' . $this->db->quoteName('disabler') . ' = ' . $parent_customfield_id
						. ' OR ' . $this->db->quoteName('override') . ' = ' . $parent_customfield_id
						. ')'
					);
				$this->db->setQuery($query, 0, $totalParent);
				$resultList = $this->db->loadRowList();

				$query->clear()
					->update($this->db->quoteName('#__virtuemart_product_customfields'));
				for ($i = 0; $i < $totalParent; $i++)
				{
					if (isset($resultList[$i][0]))
					{
						$query->clear('set')
							->clear('where')
							->set(
								$this->db->quoteName('virtuemart_product_id') . ' = ' . $this->db->quote(
									$virtuemart_product_id
								)
							);

						if (!empty($resultList[$i][1]))
						{
							$query->set($this->db->quoteName('disabler') . ' = ' . (int) $resultParent[$i][0]);
						}

						if (!empty($resultList[$i][2]))
						{
							$query->set($this->db->quoteName('override') . ' = ' . (int) $resultParent[$i][0]);
						}

						$query->where(
							$this->db->quoteName('virtuemart_customfield_id') . ' = ' . (int) $resultList[$i][0]
						);
						$this->db->setQuery($query)->execute();
						$log->add('Update disabler and override with product id');
					}
				}
			}
		}
	}

	/**
	 * Get the custom Id for customfieldsforall plugin.
	 *
	 * @param   int  $virtuemartCustomfieldId  The ID of the custom field.
	 *
	 * @return  int  Virtuemart custom fields for all id.
	 *
	 * @since   6.5.6
	 */
	private function getCf4allCustomId($virtuemartCustomfieldId)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('c.virtuemart_custom_id'))
			->from($this->db->quoteName('#__virtuemart_customs', 'c'))
			->leftJoin(
				$this->db->quoteName('#__virtuemart_product_customfields', 'pc')
				. ' ON ' . $this->db->quoteName('pc.virtuemart_custom_id') . ' = ' . $this->db->quoteName(
					'c.virtuemart_custom_id'
				)
			)
			->where($this->db->quoteName('c.custom_element') . ' = ' . $this->db->quote('customfieldsforall'))
			->where($this->db->quoteName('pc.virtuemart_customfield_id') . ' = ' . (int) $virtuemartCustomfieldId);
		$this->db->setQuery($query);

		return $this->db->loadResult();
	}

	/**
	 * Method to process custom fields for all.
	 *
	 * @param   string         $params                   An array of settings.
	 * @param   int            $virtuemartProductId      The product ID.
	 * @param   int            $virtuemartCustomId       The custom ID.
	 * @param   int            $virtuemartCustomFieldId  The custom field ID.
	 * @param   CsviHelperLog  $log                      The CSVI logger.
	 * @param   array          $multiLangValues          Multi language value array
	 * @param   string         $customParamParentValue   Parent custom value for dependent custom fields
	 *
	 * @return  mixed  Void if plugin needs to handle request | False if plugin does not handle request.
	 *
	 * @since   7.13.0
	 *
	 * @throws  Exception
	 * @throws  RuntimeException
	 */
	private function processCustomFieldsForAll(
		string $params,
		int $virtuemartProductId,
		int $virtuemartCustomId,
		int $virtuemartCustomFieldId,
		CsviHelperLog $log,
		array $multiLangValues,
		string $customParamParentValue
	) {
		$customParamValues   = explode('#', $params);
		$customParamValues   = $this->db->quote(array_unique($customParamValues));
		$multiLanguageValues = [];
		$newInsertIds        = [];

		if ($multiLangValues)
		{
			foreach ($multiLangValues as $lang => $langVal)
			{
				$multiLanguageValues[$lang] = explode('#', $langVal);
			}
		}

		[$existingInsertIds, $valueNames] = $this->processExistingValues($virtuemartCustomId, $customParamValues, $multiLanguageValues, $customParamParentValue, $log);

		$log->add('Retrieve custom field value IDs');

		$leftValueIds = array_diff($customParamValues, $valueNames);

		if ($leftValueIds)
		{
			$newInsertIds = $this->processLeftOverValues($virtuemartCustomId, $leftValueIds, $multiLanguageValues, $customParamValues, $customParamParentValue, $log);
		}

		$insertIds = array_merge($existingInsertIds, $newInsertIds);
		$this->storeValueIds($insertIds, $virtuemartProductId, $virtuemartCustomFieldId, $log);

		// Once the values are moved to custom fields table, clear them from product custom fields table
		$this->emptyCustomParams((int) $virtuemartCustomFieldId);
	}

	/**
	 * Method to get parent custom field ID
	 *
	 * @param   string         $parentCustomValue  The parent custom field name
	 * @param   CsviHelperLog  $log                The CSVI logger.
	 *
	 * @return  int ID of parent custom field value
	 *
	 * @since   7.19.0
	 */
	private function getDependentParentValue(string $parentCustomValue, CsviHelperLog $log): int
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('customsforall_value_id'))
			->from($this->db->quoteName('#__virtuemart_custom_plg_customsforall_values'))
			->where($this->db->quoteName('customsforall_value_name') . ' = ' . $this->db->quote($parentCustomValue));
		$this->db->setQuery($query);
		$log->add('Get dependent parent custom field ID');

		return $this->db->loadResult() ?? 0;
	}

	/**
	 * Method to clear the custom field values.
	 *
	 * @param   string         $plugin                 The ID of the plugin.
	 * @param   int            $virtuemart_product_id  The product ID.
	 * @param   CsviHelperLog  $log                    The CSVI logger.
	 *
	 * @return  void
	 *
	 * @since   7.2.0
	 * @throws  JsonException
	 */
	public function clearDisablerOverrideValues($plugin, $virtuemart_product_id, $log)
	{
		if (!$this->canRun)
		{
			return;
		}

		if ($plugin === 'stockablecustomfields')
		{
			$mainCustomId    = $this->getMainCustomId($virtuemart_product_id);
			$childProductIds = $this->getChildIds($mainCustomId, $virtuemart_product_id);
			$childIds        = implode(',', $childProductIds);

			if ($childIds)
			{
				$query = $this->db->getQuery(true)
					->delete($this->db->quoteName('#__virtuemart_product_customfields'))
					->where($this->db->quoteName('virtuemart_product_id') . ' IN (' . $childIds . ')')
					->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $mainCustomId);
				$this->db->setQuery($query)->execute();
				$log->add('Remove existing disabler and override values');
			}
		}
	}

	/**
	 * Get the child ids added for stockablecustomfields plugin.
	 *
	 * @param   int  $mainCustomId  The main ID of the stockablecustomfield.
	 * @param   int  $productId     The product id.
	 *
	 * @return  array  Virtuemart child id array.
	 *
	 * @since   7.2.0
	 * @throws  JsonException
	 */
	private function getChildIds($mainCustomId, $productId)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('customfield_params'))
			->from($this->db->quoteName('#__virtuemart_product_customfields'))
			->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $mainCustomId)
			->where($this->db->quoteName('virtuemart_product_id') . ' = ' . (int) $productId)
			->where($this->db->quoteName('customfield_params') . ' != ' . $this->db->quote(''));

		$this->db->setQuery($query);
		$results = $this->db->loadObjectList();
		$childId = [];

		foreach ($results as $result)
		{
			$cparams = explode('|', $result->customfield_params);

			foreach ($cparams as $cparam)
			{
				if (strpos($cparam, '='))
				{
					// Get the params
					list($name, $value) = explode('=', $cparam);

					if ($name === 'child_product_id')
					{
						$childId[] = json_decode($value, true, 512, JSON_THROW_ON_ERROR);
					}
				}
			}
		}

		return $childId;
	}

	/**
	 * Method to update override and disabler.
	 *
	 * @param   string         $plugin                     The ID of the plugin.
	 * @param   int            $virtuemart_product_id      The product ID.
	 * @param   int            $virtuemart_custom_id       The custom ID.
	 * @param   int            $virtuemart_customfield_id  The custom field ID.
	 * @param   CsviHelperLog  $log                        The CSVI logger.
	 * @param   string         $disabler                   The disabler.
	 * @param   string         $override                   The override.
	 * @param   int            $totalTitle                 The total number of custom title
	 *
	 * @return   mixed  Void if plugin needs to handle request | False if plugin does not handle request.
	 *
	 * @since   7.2.0
	 */
	public function saveDisablerOverride(
		$plugin,
		$virtuemart_product_id,
		$virtuemart_custom_id,
		$virtuemart_customfield_id,
		$log,
		$disabler,
		$override,
		$totalTitle
	) {
		if (!$this->canRun || $plugin !== 'stockablecustomfields')
		{
			return false;
		}

		if (($disabler === 'Y' || $override === 'Y'))
		{
			$disablerValue = 0;
			$overrideValue = 0;
			$pluginName    = $this->usesStockablecustomfields($virtuemart_product_id);

			$query = $this->db->getQuery(true)
				->select($this->db->quoteName('virtuemart_customfield_id'))
				->from($this->db->quoteName('#__virtuemart_product_customfields'))
				->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $virtuemart_custom_id)
				->where($this->db->quoteName('customfield_value') . ' = ' . $this->db->quote($pluginName))
				->where($this->db->quoteName('customfield_params') . ' != ' . $this->db->quote(''))
				->where($this->db->quoteName('virtuemart_product_id') . ' = ' . (int) $virtuemart_product_id)
				->where(
					'disabler = ' . $virtuemart_customfield_id . ' OR override = ' . $virtuemart_customfield_id
				);
			$this->db->setQuery($query);
			$result = $this->db->loadResult();

			if ($disabler === 'Y')
			{
				$disablerValue = $virtuemart_customfield_id;
			}

			if ($override === 'Y')
			{
				$overrideValue = $virtuemart_customfield_id;
			}

			if (!$result)
			{
				for ($i = 0; $i < $totalTitle; $i++)
				{
					$query = $this->db->getQuery(true)
						->insert($this->db->quoteName('#__virtuemart_product_customfields'))
						->columns(
							$this->db->quoteName(
								[
									'virtuemart_custom_id',
									'customfield_value',
									'disabler',
									'override',
								]
							)
						);
					$query->values(
						(int) $virtuemart_custom_id . ',' .
						$this->db->quote($pluginName) . ',' .
						$disablerValue . ',' .
						$overrideValue
					);

					$this->db->setQuery($query)->execute();
					$log->add('Create new entries for disabler and override');
				}
			}
		}
	}

	/**
	 * Export the values.
	 *
	 * @param   string         $plugin                     The ID of the plugin.
	 * @param   string         $custom_param               The custom parameter values.
	 * @param   int            $virtuemart_product_id      The product ID.
	 * @param   int            $virtuemart_custom_id       The custom ID.
	 * @param   int            $virtuemart_customfield_id  The custom field ID.
	 * @param   CsviHelperLog  $log                        The CSVI logger.
	 * @param   string         $lang                       Multi language code
	 *
	 * @return  array| void  List of values to export.
	 *
	 * @since   6.0
	 */
	public function exportCustomValues(
		$plugin,
		$custom_param,
		$virtuemart_product_id,
		$virtuemart_custom_id,
		$virtuemart_customfield_id,
		$log,
		$lang = ''
	) {
		if (!$this->canRun)
		{
			return;
		}

		if (strpos($plugin, 'customfieldsforall') !== false)
		{
			// Get the values for this custom field
			$query = $this->db->getQuery(true)
				->from($this->db->quoteName('#__virtuemart_product_custom_plg_customsforall', 'r'))
				->leftJoin(
					$this->db->quoteName('#__virtuemart_custom_plg_customsforall_values', 'v')
					. ' ON ' . $this->db->quoteName('r.customsforall_value_id') . ' = ' . $this->db->quoteName(
						'v.customsforall_value_id'
					)
				);

			$defaultLanguage = $this->virtuemartConfig->get('vmDefLang');


			if (empty($defaultLanguage))
			{
				$defaultLanguage = ComponentHelper::getParams('com_languages')->get('site');
			}

			$defaultLanguage = strtolower(str_replace('-', '_', $defaultLanguage));

			if ($lang && $lang !== $defaultLanguage)
			{
				$tableName = '#__virtuemart_custom_plg_customsforall_values_' . $lang;

				$tables = $this->db->getTableList();

				if (!in_array(str_replace('#__', $this->db->getPrefix(), $tableName), $tables, true))
				{
					throw new InvalidArgumentException(Text::sprintf('COM_CSVI_MISSING_CF4ALL_TABLE', $tableName));
				}

				$query->select(
					$this->db->quoteName('cv.customsforall_value_name', 'value') . ',' .
					$this->db->quoteName('c.virtuemart_custom_id')
				);
				$query->leftJoin(
					$this->db->quoteName($tableName, 'cv')
					. ' ON ' . $this->db->quoteName('cv.customsforall_value_id') . ' = ' . $this->db->quoteName(
						'v.customsforall_value_id'
					)
				);
			}
			else
			{
				$query->select(
					$this->db->quoteName('v.customsforall_value_name', 'value') . ',' .
					$this->db->quoteName('c.virtuemart_custom_id')
				);
			}

			$query->leftJoin(
				$this->db->quoteName('#__virtuemart_customs', 'c')
				. ' ON ' . $this->db->quoteName('v.virtuemart_custom_id') . ' = ' . $this->db->quoteName(
					'c.virtuemart_custom_id'
				)
			)
				->leftJoin(
					$this->db->quoteName('#__virtuemart_product_customfields', 'f')
					. ' ON ' . $this->db->quoteName('c.virtuemart_custom_id') . ' = ' . $this->db->quoteName(
						'f.virtuemart_custom_id'
					)
					. ' AND ' . $this->db->quoteName('r.virtuemart_product_id') . ' = ' . $this->db->quoteName(
						'f.virtuemart_product_id'
					)
				)
				->where($this->db->quoteName('r.virtuemart_product_id') . ' = ' . (int) $virtuemart_product_id)
				->where($this->db->quoteName('v.virtuemart_custom_id') . ' = ' . (int) $virtuemart_custom_id)
				->where($this->db->quoteName('r.customfield_id') . ' = ' . (int) $virtuemart_customfield_id)
				->group($this->db->quoteName('value'));
			$this->db->setQuery($query);
			$options = $this->db->loadObjectList();

			$log->add('Get CF4All values');

			// Group the data correctly
			$newoptions = [];

			foreach ($options as $option)
			{
				$newoptions[$option->virtuemart_custom_id][] = $option->value;
			}

			$values = [];

			// Create the CSVI format
			// option1[value1#value2;option2[value1#value2
			foreach ($newoptions as $option)
			{
				$values[] = implode('#', $option);
			}

			return $values;
		}

		return null;
	}

	/**
	 * Method to load plugin specific customfields.
	 *
	 * @param   CsviHelperLog  $log  The CSVI logger.
	 *
	 * @return  array|void  List of custom fields.
	 *
	 * @since   6.6.0
	 *
	 * @throws  RuntimeException
	 */
	public function loadPluginCustomFields($log)
	{
		if (!$this->canRun)
		{
			return;
		}

		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('virtuemart_custom_id', 'id'))
			->select('TRIM(' . $this->db->quoteName('custom_title') . ') AS ' . $this->db->quoteName('title'))
			->select(
				$this->db->quoteName(
					[
						'custom_parent_id',
						'field_type',
						'ordering',
					]
				)
			)
			->from($this->db->quoteName('#__virtuemart_customs'))
			->where($this->db->quoteName('custom_element') . ' = ' . $this->db->quote('customfieldsforall'));

		$this->db->setQuery($query);

		try
		{
			$customFields = $this->db->loadObjectList();
			$log->add('Get all CF4All custom field title');
		}
		catch (RuntimeException $e)
		{
			$customFields = [];
		}

		return $customFields;
	}

	/**
	 * Method to get custom field value with id.
	 *
	 * @param   string         $key  The custom parameter key.
	 * @param   int            $id   The custom ID.
	 * @param   CsviHelperLog  $log  The CSVI logger.
	 *
	 * @return  string|void  Custom field value
	 *
	 * @since   7.0
	 *
	 * @throws  RuntimeException
	 */
	public function getPluginCustomFieldValue($key, $id, $log)
	{
		if (!$this->canRun)
		{
			return;
		}

		if (strpos($key, 'customsforall_option') !== false)
		{
			$query = $this->db->getQuery(true)
				->select($this->db->quoteName('customsforall_value_name'))
				->from($this->db->quoteName('#__virtuemart_custom_plg_customsforall_values', 'cfv'))
				->leftJoin(
					$this->db->quoteName('#__virtuemart_product_custom_plg_customsforall', 'cf')
					. ' ON ' . $this->db->quoteName('cf.customsforall_value_id') . ' = ' . $this->db->quoteName(
						'cfv.customsforall_value_id'
					)
				)
				->where($this->db->quoteName('cf.id') . ' = ' . (int) $id);
			$this->db->setQuery($query);

			try
			{
				$customfieldValue = $this->db->loadResult();
				$log->add('Get CF4All custom field value');
			}
			catch (RuntimeException $e)
			{
				$customfieldValue = '';
			}

			return $customfieldValue;
		}
	}

	/**
	 * Method to check if CF4ALL plugin installed
	 *
	 * @param   string  $plugin  The plugin value
	 *
	 * @return   bool  true if plugin is installed | False if plugin not installed.
	 *
	 * @since   7.3.0
	 */
	public function checkIfPluginInstalled($plugin)
	{
		if (strpos($plugin, 'customfieldsforall') !== false || $plugin === 'stockablecustomfields')
		{
			return true;
		}

		return false;
	}

	/**
	 * Export product attribute values.
	 *
	 * @param   array   $optionValues  The product attribute values.
	 * @param   string  $fieldType     Field type
	 * @param   int     $ftype         The main custom ID.
	 *
	 * @return  mixed|void  List of values to export, False otherwise.
	 *
	 * @since   7.11.0
	 */
	public function exportProductAttribute($optionValues, $fieldType, $ftype)
	{
		if (!$this->canRun)
		{
			return;
		}

		$pluginKey = array_keys($optionValues);

		if (!in_array($fieldType[2], ['customfieldsforall', 'dependentcustomfieldsforall'], true))
		{
			return;
		}

		$pluginKeyValue = $pluginKey[0];
		$values         = '';

		// Check if it is customfieldsfor all
		if (is_array($optionValues[$pluginKeyValue]) && isset($optionValues[$pluginKeyValue]['customsforall_option']))
		{
			$customFieldsForAllId = $optionValues[$pluginKeyValue]['customsforall_option'];
			$query                = $this->db->getQuery(true)
				->select($this->db->quoteName('custom_plg_customsforall_values.customsforall_value_name'))
				->from(
					$this->db->quoteName(
						'#__virtuemart_product_custom_plg_customsforall',
						'product_custom_plg_customsforall'
					)
				)
				->leftJoin(
					$this->db->quoteName(
						'#__virtuemart_custom_plg_customsforall_values',
						'custom_plg_customsforall_values'
					)
					. ' ON ' . $this->db->quoteName(
						'custom_plg_customsforall_values.customsforall_value_id'
					) . ' = ' . $this->db->quoteName('product_custom_plg_customsforall.customsforall_value_id')
				)
				->where(
					$this->db->quoteName('product_custom_plg_customsforall.id') . ' = ' . (int) $customFieldsForAllId
				);
			$this->db->setQuery($query);
			$customFieldsForAllValue = $this->db->loadResult();
			$values                  = \JText::_($fieldType[1]) . ' - ' . $customFieldsForAllValue;
		}

		// If $values are empty then there is no CF4ALL, check if it SCF fields
		if (!$values)
		{
			$childId             = '';
			$customIds           = [];
			$stockableTitleValue = [];

			// The the custom parameters for given field type
			$query = $this->db->getQuery(true)
				->select($this->db->quoteName('custom_params'))
				->from($this->db->quoteName('#__virtuemart_customs'))
				->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $ftype);
			$this->db->setQuery($query);

			$customParams     = $this->db->loadResult();
			$mainCustomParams = explode('|', $customParams);

			foreach ($mainCustomParams as $mcParams)
			{
				if (strpos($mcParams, '='))
				{
					list($mname, $mvalue) = explode('=', $mcParams);

					if ($mname == 'custom_id')
					{
						$customIds = json_decode($mvalue);
					}
				}
			}

			$query->clear()
				->select($this->db->quoteName(['customfield_value', 'customfield_params']))
				->from($this->db->quoteName('#__virtuemart_product_customfields'))
				->where($this->db->quoteName('virtuemart_customfield_id') . ' = ' . (int) $pluginKeyValue);
			$this->db->setQuery($query);
			$customFieldsForAllValue = $this->db->loadRow();

			// Make sure both values exist
			if (!isset($customFieldsForAllValue[0]) || !isset($customFieldsForAllValue[1]))
			{
				return false;
			}

			if ($customFieldsForAllValue[0] === 'stockablecustomfields')
			{
				$cparams = explode('|', $customFieldsForAllValue[1]);

				foreach ($cparams as $cparam)
				{
					if (strpos($cparam, '='))
					{
						// Get the params
						list($name, $value) = explode('=', $cparam);

						if ($name == 'child_product_id')
						{
							$childId = json_decode($value);
						}
					}
				}

				$query = $this->db->getQuery(true)
					->select($this->db->quoteName('custom_title'))
					->from($this->db->quoteName('#__virtuemart_customs'));

				$queryCustomFields = $this->db->getQuery(true)
					->select($this->db->quoteName('customfield_value'))
					->from($this->db->quoteName('#__virtuemart_product_customfields'));

				foreach ($customIds as $cids)
				{
					$query->clear('where')
						->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $cids);
					$this->db->setQuery($query);
					$stockableTitle = $this->db->loadResult();

					$queryCustomFields->clear('where')
						->where($this->db->quoteName('virtuemart_product_id') . ' = ' . (int) $childId)
						->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . (int) $cids);
					$this->db->setQuery($query);
					$stockableValue = $this->db->loadResult();

					$stockableTitleValue[] = $stockableTitle . ' - ' . $stockableValue;
				}

				$values = \JText::_($fieldType[1]) . ' - ' . implode('#', $stockableTitleValue);
			}
		}

		return $values;
	}

	/**
	 * Load the value IDs for a given custom value and custom ID.
	 *
	 * @param   int    $virtueMartCustomId  The custom ID to filter on
	 * @param   array  $valueNames          The value names to get the IDs for
	 *
	 * @return  array  List of value IDs.
	 *
	 * @since   7.19.0
	 */
	private function getValueIds(int $virtueMartCustomId, array $valueNames): array
	{
		$db    = $this->db;
		$query = $db->getQuery(true)
			->select(
				$db->quoteName(
					[
						'customsforall_value_id',
						'customsforall_value_name',
					]
				)
			)
			->from($db->quoteName('#__virtuemart_custom_plg_customsforall_values'))
			->where($db->quoteName('virtuemart_custom_id') . ' = ' . $virtueMartCustomId)
			->where($db->quoteName('customsforall_value_name') . ' IN (' . implode(',', $valueNames) . ')');
		$db->setQuery($query);

		return $db->loadObjectList();
	}

	/**
	 * Update date the dependent parent ID.
	 *
	 * @param   int  $dependentParentId  The parent ID to set
	 * @param   int  $valueId            The value to update the parent of
	 *
	 * @return  void
	 *
	 * @since   7.19.0
	 */
	private function updateDependentParentId(int $dependentParentId, int $valueId): void
	{
		$db    = $this->db;
		$query = $db->getQuery(true)
			->update($db->quoteName('#__virtuemart_custom_plg_customsforall_values'))
			->set($db->quotename('parent_id') . ' = ' . $dependentParentId)
			->where(
				$db->quotename(
					'customsforall_value_id'
				) . ' = ' . $valueId
			);
		$db->setQuery($query)->execute();
	}

	/**
	 * Insert a new custom value.
	 *
	 * @param   string  $tableName  The table name to process
	 * @param   int     $valueId    The value ID to store
	 * @param   string  $valueName  The value name to store
	 *
	 * @return  void
	 *
	 * @since   7.19.0
	 */
	private function insertCustomValue(string $tableName, int $valueId, string $valueName): void
	{
		$db    = $this->db;
		$query = $db->getQuery(true)
			->delete($tableName)
			->where(
				$db->quoteName(
					'customsforall_value_id'
				) . ' = ' . $valueId
			);
		$db->setQuery($query)->execute();
		$query->clear()
			->insert($tableName)
			->columns(['customsforall_value_id', 'customsforall_value_name'])
			->values(
				$valueId . ',' . $db->quote($valueName)
			);
		$db->setQuery($query)->execute();
	}

	/**
	 * Get the last ordering.
	 *
	 * @param   int  $customId  The custom field ID to get the last ordering number for
	 *
	 * @return  int  The last ordering number.
	 *
	 * @since   7.19.0
	 */
	private function getLastOrdering(int $customId): int
	{
		$db    = $this->db;
		$query = $db->getQuery(true)
			->select('MAX(' . $this->db->quoteName('ordering') . ')')
			->from($this->db->quoteName('#__virtuemart_custom_plg_customsforall_values'))
			->where($this->db->quoteName('virtuemart_custom_id') . ' = ' . $customId);
		$this->db->setQuery($query);
		$lastOrdering = $this->db->loadResult();

		if (!$lastOrdering)
		{
			$lastOrdering = 0;
		}

		return (int) $lastOrdering;
	}

	/**
	 * Process the custom values.
	 *
	 * @param   int            $customId             The custom ID to process the values for
	 * @param   array          $valueIds             The values to process
	 * @param   array          $multiLanguageValues  The multi-language values to process
	 * @param   array          $valueNames           The value names to get the IDs for
	 * @param   string         $parentCustomValue    Parent custom field values
	 * @param   CsviHelperLog  $log                  The CSVI logger
	 *
	 * @return  array
	 *
	 * @since   7.19.0
	 */
	private function processLeftOverValues(
		int $customId,
		array $valueIds,
		array $multiLanguageValues,
		array $valueNames,
		string $parentCustomValue,
		CsviHelperLog $log
	): array {
		$db           = $this->db;
		$lastOrdering = $this->getLastOrdering($customId);
		$tables       = $db->getTableList();
		$prefix       = $db->getPrefix();
		$insertIds    = [];

		$newParentArray = $this->processParentValues($parentCustomValue, $valueNames);

		foreach ($valueIds as $order => $entry)
		{
			$dependentParentId = 0;
			$parentValue       = $newParentArray[$entry] ?? $parentCustomValue;

			if ($parentValue)
			{
				$dependentParentId = $this->getDependentParentValue($parentValue, $log);
			}

			$query = $db->getQuery(true)
				->insert($db->quoteName('#__virtuemart_custom_plg_customsforall_values'))
				->columns(
					[
						'customsforall_value_name',
						'virtuemart_custom_id',
						'ordering',
						'parent_id'
					]
				)
				->values(
					$entry . ','
					. $customId . ','
					. (++$lastOrdering) . ', '
					. $dependentParentId
				);
			$db->setQuery($query)
				->execute();

			$log->add('Insert new custom field value row');

			$insertId    = $db->insertid();
			$insertIds[] = $insertId;

			// Also do the multi language table entries
			if (!$multiLanguageValues)
			{
				continue;
			}

			foreach ($multiLanguageValues as $pkey => $pLangValue)
			{
				$table = $prefix . 'virtuemart_custom_plg_customsforall_values_' . strtolower(
						str_replace(
							'-',
							'_',
							$pkey
						)
					);

				if (!in_array($table, $tables, true))
				{
					$log->add('Table ' . $table . ' does not exists', false);
					$log->addStats('Skipped', Text::sprintf('COM_CSVI_CF4ALL_TABLE_DOES_NOT_EXISTS', $table));

					continue;
				}

				$tableName = $db->quoteName('#__virtuemart_custom_plg_customsforall_values_' . $pkey);
				$query->clear()
					->delete($tableName)
					->where($db->quoteName('customsforall_value_id') . ' = ' . (int) $insertId);
				$db->setQuery($query)->execute();
				$query->clear()
					->insert($tableName)
					->columns(
						[
							'customsforall_value_id',
							'customsforall_value_name'
						]
					)
					->values((int) $insertId . ',' . $db->quote($pLangValue[$order]));
				$db->setQuery($query)->execute();
			}
		}

		return $insertIds;
	}

	/**
	 * Process the parent values.
	 *
	 * @param   int            $virtueMartCustomId   The custom ID to filter on
	 * @param   array          $valueNames           The value names to get the IDs for
	 * @param   array          $multiLanguageValues  List of translated values
	 * @param   string         $parentCustomValue    Parent custom field values
	 * @param   CsviHelperLog  $log                  The CSVI logger
	 *
	 * @return  array
	 *
	 * @since   7.19.0
	 */
	private function processExistingValues(
		int $virtueMartCustomId,
		array $valueNames,
		array $multiLanguageValues,
		string $parentCustomValue,
		CsviHelperLog $log
	): array {
		$db             = $this->db;
		$valueIds       = $this->getValueIds($virtueMartCustomId, $valueNames);
		$tables         = $db->getTableList();
		$newParentArray = $this->processParentValues($parentCustomValue, $valueNames);
		$valueNames     = [];

		foreach ($valueIds as $key => $valueId)
		{
			$dependentParentId = 0;
			$parentValue       = $newParentArray[$valueId->customsforall_value_name] ?? $parentCustomValue;

			if ($parentValue)
			{
				$dependentParentId = $this->getDependentParentValue($parentValue, $log);
			}

			if ($dependentParentId)
			{
				$this->updateDependentParentId($dependentParentId, (int) $valueId->customsforall_value_id);
				$log->add('Update parent id for dependent custom field');
			}

			$valueIds[$key]   = $valueId->customsforall_value_id;
			$valueNames[$key] = $this->db->quote($valueId->customsforall_value_name);

			// Also do the multi language table entries
			if ($multiLanguageValues)
			{
				$prefix = $this->db->getPrefix();

				foreach ($multiLanguageValues as $pkey => $pLangValue)
				{
					$table = $prefix . 'virtuemart_custom_plg_customsforall_values_' . strtolower(
							str_replace('-', '_', $pkey)
						);

					if (in_array($table, $tables, true))
					{
						$tableName = $this->db->quoteName('#__virtuemart_custom_plg_customsforall_values_' . $pkey);
						$this->insertCustomValue(
							$tableName,
							(int) $valueId->customsforall_value_id,
							$pLangValue[$key]
						);
					}
					else
					{
						$log->add('Table ' . $table . ' does not exists', false);
						$log->addStats('Skipped', Text::sprintf('COM_CSVI_CF4ALL_TABLE_DOES_NOT_EXISTS', $table));
					}
				}
			}
		}

		return [$valueIds, $valueNames];
	}

	/**
	 * Rearrange parent array values so it is easy to map parent-child
	 *
	 * @param   string   $parentCustomValue  The parent custom value
	 * @param   array    $valueNames         Array of child custom values
	 *
	 * @return  array
	 *
	 * @since   7.19.0
	 */
	private function processParentValues(string $parentCustomValue, array $valueNames): array
	{
		$parentValueArray = [];

		if (strpos($parentCustomValue, '#') !== false)
		{
			$parentArray = explode('#', $parentCustomValue);

			foreach ($parentArray as $key => $parentValue)
			{
				if (isset($valueNames[$key]))
				{
					$parentValueArray[$valueNames[$key]] = $parentValue;
				}
			}
		}

		return $parentValueArray;
	}

	/**
	 * Store the value ID relations.
	 *
	 * @param   array          $valueIds       The list of IDs to store
	 * @param   int            $productId      The product ID the values belong to
	 * @param   int            $customFieldId  The custom field ID the values be long to
	 * @param   CsviHelperLog  $log            The CSVI logger
	 *
	 * @return  void
	 *
	 * @since   7.19.0
	 */
	private function storeValueIds(array $valueIds, int $productId, int $customFieldId, CsviHelperLog $log): void
	{
		if (!is_array($valueIds) || 0 === count($valueIds))
		{
			return;
		}

		$db          = $this->db;
		$deleteQuery = $db->getQuery(true)
			->delete($db->quoteName('#__virtuemart_product_custom_plg_customsforall'));
		$insertQuery = $db->getQuery(true)
			->insert($db->quoteName('#__virtuemart_product_custom_plg_customsforall'))
			->columns($db->quoteName(['customsforall_value_id', 'virtuemart_product_id', 'customfield_id']));

		foreach ($valueIds as $valueId)
		{
			$deleteQuery->clear('where')
				->where($db->quoteName('customsforall_value_id') . ' = ' . (int) $valueId)
				->where($db->quoteName('virtuemart_product_id') . ' = ' . $productId);
			$db->setQuery($deleteQuery)->execute();

			$insertQuery
				->values(
					(int) $valueId . ',' .
					$productId . ',' .
					$customFieldId
				);
		}

		$db->setQuery($insertQuery)
			->execute();

		$log->add('Insert values for custom field');
	}

	/**
	 * Moved the values to individual custom fields for all tables so clear value from product custom fields table.
	 *
	 * @param   int  $customFieldId The custom field ID to empty the parameters for
	 *
	 * @return  void
	 *
	 * @since   7.19.0
	 */
	private function emptyCustomParams(int $customFieldId): void
	{
		$db    = $this->db;
		$query = $db->getQuery(true)
			->update($db->quoteName('#__virtuemart_product_customfields'))
			->set($db->quotename('customfield_params') . ' = ' . $db->quote(''))
			->where($db->quotename('virtuemart_customfield_id') . ' = ' . $customFieldId);
		$db->setQuery($query)
			->execute();
	}

	/**
	 * Method to clear the custom field values.
	 *
	 * @param   string         $plugin                 The ID of the plugin.
	 * @param   array          $params                 An array of settings.
	 * @param   int            $virtuemartProductId    The product ID.
	 * @param   int            $virtuemartCustomId     The custom ID.
	 * @param   CsviHelperLog  $log                    The CSVI logger.
	 * @param   int            $customFieldId          The virtuemart custom field ID.
	 *
	 * @return  void
	 *
	 * @since   6.0
	 */
	public function clearCustomValues($plugin, $params, $virtuemartProductId, $virtuemartCustomId, $log, $customFieldId)
	{
		if (!$this->canRun)
		{
			return;
		}

		if (strpos($plugin, 'customfieldsforall') !== false)
		{
			$query = $this->db->getQuery(true)
				->select($this->db->quoteName('customsforall_value_id'))
				->from($this->db->quoteName('#__virtuemart_product_custom_plg_customsforall'))
				->where($this->db->quoteName('virtuemart_product_id') . ' = ' . $virtuemartProductId)
				->where($this->db->quoteName('customfield_id') . ' = ' . $customFieldId);
			$this->db->setQuery($query);
			$log->add('Select existing values from CF4ALL table');
			$existingValues = $this->db->loadColumn();

			$query->clear()
				->delete($this->db->quoteName('#__virtuemart_custom_plg_customsforall_values'))
				->where($this->db->quoteName('customsforall_value_id') . ' IN ( ' . implode(',', $existingValues) . ')');
			$this->db->setQuery($query)->execute();
			$log->add('Delete values from CF4ALL values table');

			$tables = $this->db->getTableList();
			require_once JPATH_PLUGINS . '/csviaddon/virtuemart/com_virtuemart/helper/com_virtuemart_config.php';
			$helperConfig   = new \Com_VirtuemartHelperCom_Virtuemart_Config;
			$multiLanguages = $helperConfig->get('active_languages');

			// Also do the multi language table entries deletion
			if ($multiLanguages)
			{
				$prefix = $this->db->getPrefix();

				foreach ($multiLanguages as $multiLanguage)
				{
					$table = $prefix . 'virtuemart_custom_plg_customsforall_values_' . $multiLanguage;

					if (in_array($table, $tables, true))
					{
						$tableName = $this->db->quoteName('#__virtuemart_custom_plg_customsforall_values_' . $multiLanguage);
						$query->clear()
							->delete($tableName)
							->where($this->db->quoteName('customsforall_value_id') . ' IN ( ' . implode(',', $existingValues) . ')');
						$this->db->setQuery($query)->execute();
						$log->add('Delete values from CF4ALL values language table');
					}
					else
					{
						$log->add('Table ' . $table . ' does not exists to delete existing records', false);
						$log->addStats('Skipped', Text::sprintf('COM_CSVI_CF4ALL_TABLE_DOES_NOT_EXISTS', $table));
					}
				}
			}

			$query->clear()
				->delete($this->db->quoteName('#__virtuemart_product_custom_plg_customsforall'))
				->where($this->db->quoteName('customsforall_value_id') . ' IN ( ' . implode(',', $existingValues) . ')');
			$this->db->setQuery($query)->execute();
			$log->add('Delete values from CF4ALL table');
		}
	}
}
